/*
 * dr.h --
 *
 *  Last Modified: Tue Dec 20 18:03:29 1994 by Bryan M. Kramer
 *
 * Copyright (c) 1990-1994 University of Toronto
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF TORONTO BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * TORONTO HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF TORONTO SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF TORONTO HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

# ifndef TK_DR_H
# define TK_DR_H

# include <tcl.h>
# include "kb.h"

/* note: we are overloading KBList and will cast the result of first */

# define VARIABLE(x) ((Variable*)(x)->object)
# define RULE(x) ((Rule*)(x)->object)

# define Atom      AAtom


typedef struct s_Variable
{
    int id;
    char* name;
    Boolean list;
    KBList* types;
} Variable;


typedef struct s_Term
{
    Boolean variable;
    Boolean list;
    union
    {
	KBList*   objects;
	Variable* variable;
    } value;
} Term;


typedef struct s_Atom
{
    char* predicate;
    int   order;
    Boolean not;
    Tcl_HashEntry* entry;
    Term** terms;
} Atom;





typedef struct s_Goal
{
    int    id;
    Atom*  atom;
    struct s_Goal* next;
} Goal;


typedef struct s_Bindings
{
    Variable* variable;
    int    level;
    Term*  value;
    int    vlevel;
    struct s_Bindings* next;
    KBList* types;
} Bindings;



typedef int (*fn_defn) ();

typedef struct s_Rule
{
    char*  name;
    Atom*  head;
    Atom** tail;
    int	   tail_count;
    fn_defn prim;
} Rule;


typedef struct s_TCLInfo
{
    Tcl_Interp* interp;
    Tcl_DString result;
    int		status;
    Boolean	not;
    Boolean	succeed_once;
    Boolean	succeeded;
} TCLInfo;

typedef KBList VarList;
typedef KBList RuleList;


Atom*	new_atom ();
Term*	new_object_term ();
Term*	new_var_term ();
Term*	new_list_term ();

Variable* new_simple_var ();

void	new_rule ();
void	new_goals ();

# endif
