/*
 * kb.h --
 *
 *  Last Modified: Mon Sep 18 16:45:43 1995 by Bryan M. Kramer
 *
 * Copyright (c) 1990-1994 University of Toronto
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF TORONTO BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * TORONTO HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF TORONTO SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF TORONTO HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/* WINGS BEGIN */
#ifdef WINGS_WIN95
#include <tkWinPort.h>
#endif
/* WINGS END */
# ifndef TK_KB_H
# define TK_KB_H

# define streq(a,b)  (strcmp((a),(b))==0)
/* WINGS BEGIN */
#ifdef WINGS_UNIX
# define strcaseeq(a,b)  (strcasecmp((a),(b))==0)
#endif
#ifdef WINGS_WIN95
#define strcaseeq(a,b) (strcmpi((a),(b))==0)
#endif
/* WINGS END */



# define labeleq(a,b)  (((a)==(b))||((a) && (b) && strcmp((a),(b))==0))
# define labelcaseeq(a,b)  (((a)==(b))||((a) && (b) && strcasecmp((a),(b))==0))


# define pushf(obj,place)  ((place) = cons ((obj),(place)))
# define push(obj,place)  (list_find((obj),(place)) ? nil : ((place) = cons ((obj),(place))))
# define remove(obj,place)  ((place) = list_remove ((obj),(place)))

# define PROP  "Proposition"

# define UNDEFINED 0x1
# define ATTRIBUTE 0x2
# define TDS_DEFINITION 0x4
# define TDS_REQUESTED 0x8
# define TDS_UNKNOWN 0x10
# define LOCAL_DEF 0x20
# define TDS_INST_REQ 0x40
# define TDS_CHILDREN_REQ 0x80
# define TDS_INST_RECVD 0x100
# define TDS_CHILDREN_RECVD 0x200
# define USER_MODIFIED 0x400

/* we use the top byte of the status word for level information */

# define L_OMEGA  0xF000
# define L_TOKEN  0xE000
# define L_SCLASS  0xD000
# define L_M1  0xC000
# define L_M2  0xB000
# define L_M3  0xA000
# define L_M4  0x9000
# define L_ATTRIBUTE  0x8000
# define L_UNKNOWN 0x0000

# define LEVELMASK 0xF000
# define LEVELOF(obj) ((obj)->status & LEVELMASK)
# define SETLEVEL(obj,lvl) ((obj)->status = (~LEVELMASK & (obj)->status) | lvl)

# define STATUS(obj,s) ((Boolean) ((obj)->status & s))

typedef int Boolean;
# define true 1
# define false 0


struct s_KBObject;
# define new(type)	(type*) malloc (sizeof (type))
# define nil 0


# define SOURCE_NONE 0
# define SOURCE_FILE 1
# define SOURCE_TDS 2
# define SOURCE_TCL 3
# define SOURCE_BUILTIN 4
# define SOURCE_CKB 5


typedef struct Source_TDS
{
    char* kb_name;
} Source_TDS;

typedef struct Source_File
{
    char* name;
} Source_File;

typedef struct Source_CKB
{
    char* name;
} Source_CKB;

typedef struct Source
{
    int source_type;
    union {
	struct Source_TDS tds;
	struct Source_File file;
	struct Source_CKB ckb;
    } source;
    struct Source* next;
} Source;


extern Source* current_source;
extern Source* default_source;

typedef struct s_KBList
{
    struct s_KBObject* object;
    struct s_KBList* next;
} KBList;

typedef struct s_KBObject
{
    int seen;
    int status;
    KBList* inst_of;
    char* id;
    KBList* isa;
    KBList* instances;
    KBList* children;
    KBList* attributes;
    struct s_KBObject* attribute_of;
    KBList* back_ptr;
    struct s_KBObject* value;
    char* name;
    Source* source;
    void* client_data;
} KBObject;

KBObject* new_object ();
KBList* minimize_classes ();
KBList* list_find ();
KBList* cons ();
void free_list ();
KBList* intersect_type_lists ();
Boolean is_instance_of ();
KBList* resolve_cat_class ();
KBList* resolve_cat ();
extern int parse_file ();
void create_attr ();
int list_length ();

# endif
