/* */
/* Implement functions for talking to Telos Repository via mbus */
/* */
/* Last Modified: Fri Aug 11 11:28:07 1995 by Bryan M. Kramer
								    */
/*         Copyright (c) 1995  University of Toronto, Ontario Hydro           */
/* ======================================================================     */
/* Permission to use, copy, modify, and distribute this software and its      */
/* documentation for any non-commercial purpose is hereby granted,            */
/* provided that the above copyright notice appear in all copies and that     */
/* both that the copyright notice and warranty disclaimer appear in           */
/* supporting documentation, and that the names of the University of Toronto, */
/* Ontario Hydro, and Bryan M. Kramer not be used in advertising or publicity */
/* pertaining to distribution of the software without specific, written prior */
/* permission.                        					      */
/*                                                                            */
/* The University of Toronto, Ontario Hydro, and Bryan M. Kramer disclaim all */
/* warranties with regard to this software, including all implied warranties  */
/* of merchantability and fitness.  In no event shall the University of Toronto, */
/* Ontario Hydro, or Bryan M. Kramer be liable for any special, indirect or   */
/* consequential damages or any damages whatsoever resulting from loss of use, */
/* data or profits, whether in an action of contract, negligence or other     */
/* tortuous action, arising out of or in connection with the use or           */
/* performance of this software.                                              */
/*                                                                            */
/* */

# ifdef MBUS
# include	<config.h>
# endif

# define _NO_PROTO

# include	<fcntl.h>
# include	<stdio.h>
# include	<sys/types.h>
/* WINGS BEGIN */
#ifdef WINGS_UNIX
# include	<sys/socket.h>
# include	<netinet/in.h>

# include	<sys/errno.h>
#endif
/* WINGS END */

# ifdef MBUS
# include	<mbus/mbus.h>
# include	<mbus/api.h>
# include	<mbus/mbus.h>
# include	<mbus/api.h>
# include	<mbus/keyword.h>
# include	<cb-defs.h>
# endif

# include	<signal.h>
# include	<ctype.h>
/* WINGS BEGIN */
#ifdef WINGS_UNIX
# include	<sys/time.h>
# include	<sys/resource.h>
#endif
/* WINGS END */

# define    MILLION 	1000000

# ifndef nil
# define nil 0
# endif

# define COM_SIZE 100
# define ARG_SIZE 100

typedef int (tm_fn) ();

typedef int Boolean;
# define true 1
# define false 0

typedef struct s_TelosCmdData
{
    char* name;
    char* cmd_type;
    tm_fn* command;
    tm_fn* error_command;
} TelosCmdData;


typedef struct s_TelosRequest
{
    char id[100];
    TelosCmdData* command;
    char arg[ARG_SIZE];
    void* interp;
    void* client_data;
    char* callback;
    struct s_TelosRequest* next;
} TelosRequest;

static TelosRequest* all_requests;



int MBLogLevel = 0;

static char hello_message [] = "(id \"File Monitor\")";
static char accept_message [] = "(accept \".*\" \".\")";

# ifdef MBUS
static t_mb_connection c;
# endif

int tm_default_result ();
static char domain[1000];
static int message_count;
int tds_connection_exists;

# ifdef MBUS

int tmb (host, port)
    char*   host;
    int	    port;
{    
    char	buffer[1000];
    int pid = getpid ();
    int	conn_fd;

    all_requests = 0;

    sprintf (domain, "%s.%d", getlogin (), pid); 

    if (NULL == (c = MBConnect (host, port)))
    {
	fprintf (stderr, "Couldn't connect to MBus at %s, %d\n", 
		NULL == host ? "*empty*" : host, port);
	return 0;
    }


    conn_fd = MBconnection_fd (c);
    MBtransmit_buffer (c, hello_message, strlen (hello_message));
    sprintf (buffer, "(accept \".*\" \"RIGI.%s\")", domain);
    /* MBtransmit_buffer (c, accept_message, strlen (accept_message)); */
    MBtransmit_buffer (c, buffer, strlen (buffer));

    message_count = 0;
    tds_connection_exists = 1;

    return conn_fd;
}

int tds_connection ()
{
    return (int) (c != 0);
}

TelosRequest* find_and_remove_request (id)
    t_sexp* id;
{
    TelosRequest* prev = 0;
    TelosRequest* ptr;

    for (ptr = all_requests; ptr; prev = ptr, ptr = ptr->next)
    {
	if (MBcompare_Cstring (id, ptr->id) == 0)
	{
	    if (prev)
		prev->next = ptr->next;
	    else
		all_requests = ptr->next;

	    return ptr;
	}
    }
    return nil;
}

extern void* get_tm_command ();
extern char* strdup ();

/* TODO: make all data, result, error accumulation and error status information */
/* part of the request								 */

TelosRequest* new_request (client_data, interp, command, arg, callback, require_known)
    void*   client_data;
    void*   interp;
    char*   command;
    char*   arg;
    char*   callback;
    Boolean require_known;
{
    TelosRequest* r;
    TelosCmdData* cmd_data;

    cmd_data = (TelosCmdData*) get_tm_command (command);

    if (require_known && ! cmd_data)
	return nil;

    r = (TelosRequest*) malloc (sizeof (TelosRequest));
    r->next = all_requests;
    all_requests = r;
    r->interp = interp;
    r->client_data = client_data;
    sprintf (r->id, "%s.%d", domain, message_count);
    message_count++;
    r->command = cmd_data;
    strncpy (r->arg, arg, ARG_SIZE);
    r->arg[ARG_SIZE - 1] = 0;
    if (callback)
	r->callback = strdup (callback);
    else
	r->callback = nil;

    return r;
}



void RepInProc (clientData, mask)
    char*   clientData;
    int	    mask;
{
    t_sexp    message;
    t_sexp    stuff;
    t_sexp    id_cell;
    t_sexp    status;
    t_sexp    result;

    TelosRequest* req;
    char*	text;
    char*	status_string;

    while ((message = MBNextMessage (c)))
    {
	stuff = MBnth (message, 2);
	id_cell = MB_CAR (MBnth (stuff, 4));
	req = find_and_remove_request (id_cell);
	if (req)
	{
	    status = MB_CAR (MBnth (stuff, 5));
	    status_string = MBprint_Cstring (status);
	    if (strcmp (status_string, "0"))
	    {
		if (req->callback)
		{
		    text = MBprint_Cstring (MB_CAR (MBnth (stuff, 7)));
		    kb_call_back (req->client_data, req->interp, req->callback, 1, text);
		    free (text);
		}
		else if (req->command && req->command->error_command)
		{		    
		    (req->command->error_command) (MB_CAR (MBnth (stuff, 7)), req);
		}
		else
		{
		    text = MBprint_Cstring (MB_CAR (MBnth (stuff, 7)));
		    rep_error ("Error: ", text);
		    free (text);
		}
	    }
	    else
	    {
		result = MB_CAR (MBnth (stuff, 6));
	    
		
		if (req->callback)
		{
		    text = MBprint_Cstring (result);
		    kb_call_back (req->client_data, req->interp, req->callback, 0, text);
		    free (text);
		}
		else if (req->command && req->command->command)
		    (req->command->command) (result, req);
		else
		    tm_default_result (result, req);
	    }

	    free (status_string);
	    free (req);
	}
	MBfree (message);
    }
    
    if (! all_requests)
	rep_end_wait ();
}

static char message[8096];

void RepTell (cd, i, args, callback)
    void*   cd;
    void*   i;
    char*   args;
    char*   callback;
{
    TelosRequest* req;

    if (! c)
    {
	rep_error ("Error: no connection to mbus", 0);
	return;
    }

    req = new_request (cd, i, "tell", args, callback, false);
    sprintf (message, "(\"my-tag\" (\"TELOS_DATA_SERVER\") ((%s) ((() (RIGI) (%s))) (((() (TELOS_DATA_SERVER) ()))) (2) (TELL) ((%s))))",
	     req->id,
	     domain,
	     args);
    MBtransmit_buffer (c, message, sizeof (message) -1);
    if (! callback) 
	rep_wait (cd, i);   
}



void RepAsk (cd, i, command, names, callback)
    void*   cd;
    void*   i;
    char*   command;
    char*   names;
    char*   callback;
{
    TelosRequest* req;

    if (! c)
    {
	rep_error ("Error: no connection to mbus", 0);
	return;
    }

    req = new_request (cd, i, command, names, callback, false);
    sprintf (message, "(\"my-tag\" (\"TELOS_DATA_SERVER\") ((%s) ((() (RIGI) (%s))) (((() (TELOS_DATA_SERVER) ()))) (2) (ASK) ((%s %s))))",
	     req->id,
	     domain,
	     command,
	     names);
    MBtransmit_buffer (c, message, sizeof (message) -1);
    if (! callback) 
	rep_wait (cd, i);   
}


void RepRequestObject (cd, i, names, callback)
    void*   cd;
    void*   i;
    char*   names;
    char*   callback;
{
    TelosRequest* req;

    if (! c)
    {
	rep_error ("Error: no connection to mbus", 0);
	return;
    }

    req = new_request (cd, i, "getObject", names, callback, false);
    sprintf (message, "(\"my-tag\" (\"TELOS_DATA_SERVER\") ((%s) ((() (RIGI) (%s))) (((() (TELOS_DATA_SERVER) ()))) (2) (ASK) ((getObject %s))))",
	     req->id,
	     domain,
	     names);
    MBtransmit_buffer (c, message, sizeof (message) -1);
    if (! callback)
	rep_wait (cd, i);
}


void RepCommand (cd, i, command, name, callback)
    void*   cd;
    void*   i;
    char*   command;
    char*   name;
    char*   callback;
{
    TelosRequest* req;

    if (! c)
    {
	rep_error ("Error: no connection to mbus", 0);
	return;
    }

    req = new_request (cd, i, command, name, callback, false);
    sprintf (message, "(\"my-tag\" (\"TELOS_DATA_SERVER\") ((%s) ((() (RIGI) (%s))) (((() (TELOS_DATA_SERVER) ()))) (2) (COMMAND) ((%s %s))))",
	     req->id,
	     domain,
	     command,
	     name);
    MBtransmit_buffer (c, message, sizeof (message) -1);
    if (! callback)
	rep_wait (cd, i);
}

/* return 1 if known repository communication, 0 if not */

int  RepGeneric (cd, i, command, args, callback)
    void*   cd;
    void*   i;
    char*   command;
    char*   args;
    char*   callback;
{
    TelosRequest* req;

    if (! c)
    {
	rep_error ("Error: no connection to mbus", 0);
	return;
    }
    if (strcmp (command, "tell") == 0 || strcmp (command, "TELL") == 0)
    {
	RepTell (cd, i, args, callback);
	return;
    }

    req = new_request (cd, i, command, args, callback, true);
    if (! req)
	return 0;

    sprintf (message, "(\"my-tag\" (\"TELOS_DATA_SERVER\") ((%s) ((() (RIGI) (%s))) (((() (TELOS_DATA_SERVER) ()))) (2) (%s) ((%s %s))))",
	     req->id,
	     domain,
	     req->command->cmd_type,
	     command,
	     args);
    MBtransmit_buffer (c, message, sizeof (message) -1);
    if (! callback)
	rep_wait (cd, i);
}


/* functions called when a getObject request is responded too */

int tm_get_object (exp, req)
    t_sexp exp;
    TelosRequest* req;
{
    char* str;
    t_sexp next;

    for (next = exp; next; next = MB_CDR (next))
    {
	str = MBprint_Cstring (MB_CAR (next));    
	parse_string (str);
	free (str);
    }
    /* change defined status of requested objects */
    kb_update_status (req->client_data, req->interp, req->arg, 1);
    rep_append_result (req->interp, req->arg);
}


int tm_get_object_error (error_exp, req)
    t_sexp error_exp;
    TelosRequest* req;
{
    char* str;
    t_sexp next;
    
    /* remove requested mark from requested objects (except error object) */
    kb_update_status (req->client_data, req->interp, req->arg, 0);
    str = MBprint_Cstring (error_exp); 
    rep_error ("TDS couldn't define", str);
    free (str);

    for (next = error_exp; next; next = MB_CDR (next))
    {
	str = MBprint_Cstring (MB_CAR (next));
	rep_set_tds_error (str);
	free (str);
    }
}

/* functions called when  get names requests are responed to */


int tm_get_instance_names (exp, req)
    t_sexp exp;
    TelosRequest* req;
{
    char* str;
    t_sexp next;

    reset_child_list ();
    for (next = exp; next; next = MB_CDR (next))
    {
	str = MBprint_Cstring (MB_CAR (next));    
	rep_append_result (req->interp, str);
	rep_add_child (str);
	free (str);
    }
    rep_set_instances (req->client_data, req->interp, req->arg);
}

int tm_get_children_names (exp, req)
    t_sexp exp;
    TelosRequest* req;
{
    char* str;
    t_sexp next;

    reset_child_list ();
    for (next = exp; next; next = MB_CDR (next))
    {
	str = MBprint_Cstring (MB_CAR (next));    
	rep_append_result (req->interp, str);
	rep_add_child (str);
	free (str);
    }
    rep_set_children (req->client_data, req->interp, req->arg);
}

/* default result processor */

int tm_default_result (exp, req)
    t_sexp exp;
    TelosRequest* req;
{
    char* str;
    t_sexp next;

    for (next = exp; next; next = MB_CDR (next))
    {
	str = MBprint_Cstring (MB_CAR (next));    
	rep_append_result (req->interp, str);
	free (str);
    }
}

/* callback for boolean results */

int tm_boole_result (exp, req)
    t_sexp exp;
    TelosRequest* req;
{
    rep_append_result (req->interp, "1");
}


/* initialize callbacks */

TelosCmdData* new_cmd (fn, error_fn)
    tm_fn* fn;
    tm_fn* error_fn;
{
    TelosCmdData* c;

    c = (TelosCmdData*) malloc (sizeof (TelosCmdData));
    c->command = fn;
    c->error_command = error_fn;

    return c;
}

TelosCmdData all_telos_commands[] =
{
    {"exists", "ASK", tm_boole_result, nil},
    {"getObject", "ASK", tm_get_object, tm_get_object_error},
    {"getInstanceNames", "ASK", tm_get_instance_names, nil},
    {"getInstances", "ASK", nil, nil},
    {"getAncestorNames", "ASK", nil, nil},
    {"getAncestors", "ASK", nil, nil},
    {"getChildrenNames", "ASK", tm_get_children_names, nil},
    {"getChildren", "ASK", nil, nil},
    {"getParentNames", "ASK", nil, nil},
    {"getParents", "ASK", nil, nil},
    {"setKB", "COMMAND", nil, nil},
    {"getKB", "COMMAND", nil, nil},
    {"listKBs", "COMMAND", nil, nil},
    {"size", "COMMAND", nil, nil},
    {"deleteKB", "COMMAND", nil, nil},
    {"logon", "COMMAND", nil, nil},
    {"logoff", "COMMAND", nil, nil},
    {"parse", "COMMAND", nil, nil},
    {"tell", "TELL", nil, nil},
    {"TELL", "TELL", nil, nil},
    {nil, nil, nil, nil}
};


void tm_init ()
{
    TelosCmdData* ptr;
    tds_connection_exists = 0;

    for (ptr = all_telos_commands; ptr->name; ptr++)
    {
	register_tm_command (ptr->name, ptr);
    }
}


# else


int tmb (host, port)
    char*   host;
    int	    port;
{
    return 0;
}


int tds_connection ()
{
    return 0;
}


TelosRequest* find_and_remove_request (id)
    void* id;
{
    return 0;
}


TelosRequest* new_request (client_data, interp, command, arg, callback, require_known)
    void*   client_data;
    void*   interp;
    char*   command;
    char*   arg;
    char*   callback;
    Boolean require_known;
{
    return 0;
}



void RepInProc (clientData, mask)
    char*   clientData;
    int	    mask;
{
}



void RepTell (cd, i, args, callback)
    void*   cd;
    void*   i;
    char*   args;
    char*   callback;
{
}


void RepAsk (cd, i, command, names, callback)
    void*   cd;
    void*   i;
    char*   command;
    char*   names;
    char*   callback;
{
}



void RepRequestObject (cd, i, names, callback)
    void*   cd;
    void*   i;
    char*   names;
    char*   callback;
{
}


void RepCommand (cd, i, command, name, callback)
    void*   cd;
    void*   i;
    char*   command;
    char*   name;
    char*   callback;
{
}


int  RepGeneric (cd, i, command, args, callback)
    void*   cd;
    void*   i;
    char*   command;
    char*   args;
    char*   callback;
{
    return 0;
}


int tm_get_object (exp, req)
    void* exp;
    TelosRequest* req;
{
    return 0;
}



int tm_get_object_error (error_exp, req)
    void* error_exp;
    TelosRequest* req;
{
}

int tm_get_instance_names (exp, req)
    void* exp;
    TelosRequest* req;
{
    return 0;
}

int tm_get_children_names (exp, req)
    void* exp;
    TelosRequest* req;
{
    return 0;
}



int tm_default_result (exp, req)
    void* exp;
    TelosRequest* req;
{
    return 0;
}


int tm_boole_result (exp, req)
    void* exp;
    TelosRequest* req;
{
    return 0;
}



TelosCmdData* new_cmd (fn, error_fn)
    tm_fn* fn;
    tm_fn* error_fn;
{
    return 0;
}

/* WINGS BEGIN */
/*TelosCmdData all_telos_commands[]; */
TelosCmdData all_telos_commands[1];
/* WINGS END */


void tm_init ()
{
}

# endif
